﻿/*
 * Copyright YongChul Kwon 2009
 * 
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using LinqToDryad;

namespace dFoF.Common
{
    public struct Util
    {
        static public Char[] DELIMITOR = { ',' };
        static public string DryadLinqRootDir
        {
            get { return Environment.GetEnvironmentVariable("DRYADLINQROOT"); }
        }


        //public delegate T AverageFunc<T>(T a, T b);
        public delegate V ValueSelector<T, V>(T o);
        public delegate bool LessThanEqualTo<T>(T a, T b);

        public static void swap<T>(ref T[] arr, int a, int b)
        {
            T tmp = arr[a];
            arr[a] = arr[b];
            arr[b] = tmp;
        }

        public static int Partition<T, V>(ref T[] arr, int left, int right, int pivot, ValueSelector<T, V> val, LessThanEqualTo<V> lt)
        {
            V pivotValue = val(arr[pivot]);
            swap(ref arr, pivot, right);
            int storeIdx = left;
            for (int i = left; i < right; ++i)
            {
                if (lt(val(arr[i]), pivotValue))
                {
                    swap(ref arr, i, storeIdx);
                    ++storeIdx;
                }
            }
            swap(ref arr, storeIdx, right);
            return storeIdx;
        }

        public static int FindMax<T, V>(ref T[] arr, int left, int right, ValueSelector<T, V> val, LessThanEqualTo<V> lt)
        {
            int maxIdx = left;
            V maxVal = val(arr[maxIdx]);
            for (int i = left; i <= right; ++i)
            {
                if (lt(maxVal, val(arr[i])))
                {
                    maxIdx = i;
                    maxVal = val(arr[i]);
                }
            }
            swap(ref arr, maxIdx, right);
            return right;
        }

        public static int FindMin<T, V>(ref T[] arr, int left, int right, ValueSelector<T, V> val, LessThanEqualTo<V> lt)
        {
            int minIdx = left;
            V minVal = val(arr[minIdx]);
            for (int i = left; i <= right; ++i)
            {
                if ( lt(val(arr[i]),minVal) )
                {
                    minIdx = i;
                    minVal = val(arr[i]);
                }
            }
            swap(ref arr, minIdx, left);
            return left;
        }

        public static int FindKMedian<T, V>(int k, ref T[] e, int left, int right, Random rand, ValueSelector<T, V> val, LessThanEqualTo<V> lt)
        {
            int pidx = rand.Next(left, right+1);
            V pv = val(e[pidx]);
            int p = Partition(ref e, left, right, pidx, val, lt);
            int rank = p - left + 1;

            //Console.WriteLine("K={0},PV={5},L={1},P={2},R={3},RK={4}", k, left, p, right, rank,pv);

            //if (left > right)
            //{
            //    Console.WriteLine("Should NOT happen!");
            //    return -1;
            //}

            if ( rank == k )
            {
                return p;
            }
            else if ( rank < k )
            {
                return FindKMedian( k - rank,ref e, p + 1, right, rand, val, lt);
            }
            else
            {
                return FindKMedian(k, ref e, left, p - 1, rand, val, lt);
            }
        }

        public static V[] FindMedian<T, V>(ref T[] e, Random rand, ValueSelector<T, V> val, LessThanEqualTo<V> lt)
        {
            return FindMedian(ref e, 0, e.Length - 1, rand, val, lt);
        }

        public static V[] FindMedian<T, V>(ref T[] e, int l, int r, Random rand, ValueSelector<T, V> val, LessThanEqualTo<V> lt)
        {
            int len = r - l + 1;
            if ( (len & 0x01) == 0 ) {
                int m1 = FindKMedian(len>>1,ref e,l,r,rand,val,lt);
                //int m2 = FindKMedian(1,ref e,m1+1,r,rand,val,lt);
                int m2 = FindMin(ref e, m1 + 1, r, val, lt);
                return new V[] { val(e[m1]), val(e[m2]) };
            } else {
                // odd number
                int p = FindKMedian((len+1)>>1,ref e,l,r,rand,val,lt);
                return new V[] { val(e[p]) };
            }
        }

        public static int FindKMedian<T, V>(int k, ref T[] e, Random rand, ValueSelector<T, V> val, LessThanEqualTo<V> lt)
        {
            return FindKMedian(k, ref e, 0, e.Length - 1, rand, val, lt);
        }

        public static int FindKMedian<T, V>(int k, ref T[] e, ValueSelector<T, V> val, LessThanEqualTo<V> lt)
        {
            return FindKMedian(k, ref e, 0, e.Length - 1, new Random(), val, lt);
        }
    }

    public enum ParticleType { UNKNOWN, DARK, GAS, STAR };

    [Serializable]
    public class Particle : IEquatable<Particle>, IComparable<Particle>
    {
        public long pid;
        public float mass;
        public float x;
        public float y;
        public float z;
        public float vx;
        public float vy;
        public float vz;
        public ParticleType type;

        public long ParticleID { get { return pid; } set { pid = value; } }
        public float Mass { get { return mass; } set { mass = value; } }
        public float X { get { return x; } set { x = value; } }
        public float Y { get { return y; } set { y = value; } }
        public float Z { get { return z; } set { z = value; } }
        public float Vx { get { return vx; } set { vx = value; } }
        public float Vy { get { return vy; } set { vy = value; } }
        public float Vz { get { return vz; } set { vz = value; } }
        public ParticleType Type { get { return type; } set { type = value; } }

        public Particle(float x1, float y1, float z1)
        {
            pid = -1;
            mass = 0.0f;
            x = x1;
            y = y1;
            z = z1;
            vx = vy = vz = 0.0f;
            type = ParticleType.UNKNOWN;
        }


        public Particle(ParticleType t,long p, float m, float x1, float y1, float z1, float vx1, float vy1, float vz1)
        {
            pid = p;
            mass = m;
            x = x1;
            y = y1;
            z = z1;
            vx = vx1;
            vy = vy1;
            vz = vz1;
            type = t;
        }

        public double distance(Particle o)
        {
            float dx = x - o.X;
            float dy = y - o.Y;
            float dz = z - o.Z;
            return Math.Sqrt(dx * dx + dy * dy + dz * dz);
        }

        #region IEquatable<Particle> Members

        public bool Equals(Particle other)
        {
            return pid == other.ParticleID;
        }

        #endregion

        #region IComparable<Particle> Members

        public int CompareTo(Particle other)
        {
            return (int)(pid - other.ParticleID);
        }

        #endregion
    }

        public class BGasParticle : IEquatable<BGasParticle>, IComparable<BGasParticle>
        {
            public long pid;
            public float mass;
            public float x;
            public float y;
            public float z;
            public float vx;
            public float vy;
            public float vz;
            public float phi;
            public float rho;
            public float temp;
            public float hsmooth;
            public float metal;

            public long ParticleID { get { return pid; } set { pid = value; } }
            public float Mass { get { return mass; } set { mass = value; } }
            public float X { get { return x; } set { x = value; } }
            public float Y { get { return y; } set { y = value; } }
            public float Z { get { return z; } set { z = value; } }
            public float Vx { get { return vx; } set { vx = value; } }
            public float Vy { get { return vy; } set { vy = value; } }
            public float Vz { get { return vz; } set { vz = value; } }
            public float Phi { get { return phi; } set { phi = value; } }
            public float Rho { get { return rho; } set { rho = value; } }
            public float Temp { get { return temp; } set { temp = value; } }
            public float Hsmooth { get { return hsmooth; } set { hsmooth = value; } }
            public float Metal { get { return metal; } set { metal = value; } }

            public static BGasParticle Read(DryadBinaryReader rd)
            {
                BGasParticle p = new BGasParticle();
                p.ParticleID = rd.ReadInt64();
                p.Mass = rd.ReadSingle();
                p.X = rd.ReadSingle();
                p.Y = rd.ReadSingle();
                p.Z = rd.ReadSingle();
                p.Vx = rd.ReadSingle();
                p.Vy = rd.ReadSingle();
                p.Vz = rd.ReadSingle();
                p.Phi = rd.ReadSingle();
                p.Rho = rd.ReadSingle();
                p.Temp = rd.ReadSingle();
                p.Hsmooth = rd.ReadSingle();
                p.Metal = rd.ReadSingle();
                return p;
            }

            public static int Write(DryadBinaryWriter wr, BGasParticle p)
            {
                int l = 0;
                l += wr.Write(p.ParticleID);
                l += wr.Write(p.Mass);
                l += wr.Write(p.X);
                l += wr.Write(p.Y);
                l += wr.Write(p.Z);
                l += wr.Write(p.Vx);
                l += wr.Write(p.Vy);
                l += wr.Write(p.Vz);
                l += wr.Write(p.Phi);
                l += wr.Write(p.Rho);
                l += wr.Write(p.Temp);
                l += wr.Write(p.Hsmooth);
                l += wr.Write(p.Metal);
                return l;
            }

            public CParticle toCParticle()
            {
                return new CParticle(pid, X, Y, Z);
            }

            public Particle toParticle()
            {
                return new Particle(ParticleType.GAS, pid, mass, x, y, z, vx, vy, vz);
            }

            #region IEquatable<BGasParticle> Members

            public bool Equals(BGasParticle other)
            {
                return ParticleID == other.ParticleID;
            }

            #endregion

            #region IComparable<Particle> Members

            public int CompareTo(BGasParticle other)
            {
                return (int)(ParticleID - other.ParticleID);
            }

            #endregion
        }

        public class BDarkParticle : IEquatable<BDarkParticle>, IComparable<BDarkParticle>
        {
            public long pid;
            public float mass;
            public float x;
            public float y;
            public float z;
            public float vx;
            public float vy;
            public float vz;
            public float phi;
            public float eps;

            public long ParticleID { get { return pid; } set { pid = value; } }
            public float Mass { get { return mass; } set { mass = value; } }
            public float X { get { return x; } set { x = value; } }
            public float Y { get { return y; } set { y = value; } }
            public float Z { get { return z; } set { z = value; } }
            public float Vx { get { return vx; } set { vx = value; } }
            public float Vy { get { return vy; } set { vy = value; } }
            public float Vz { get { return vz; } set { vz = value; } }
            public float Phi { get { return phi; } set { phi = value; } }
            public float Eps { get { return eps; } set { eps = value; } }

            public static BDarkParticle Read(DryadBinaryReader rd)
            {
                BDarkParticle p = new BDarkParticle();
                p.ParticleID = rd.ReadInt64();
                p.Mass = rd.ReadSingle();
                p.X = rd.ReadSingle();
                p.Y = rd.ReadSingle();
                p.Z = rd.ReadSingle();
                p.Vx = rd.ReadSingle();
                p.Vy = rd.ReadSingle();
                p.Vz = rd.ReadSingle();
                p.Phi = rd.ReadSingle();
                p.Eps = rd.ReadSingle();
                return p;
            }

            public static int Write(DryadBinaryWriter wr, BDarkParticle p)
            {
                int l = 0;
                l += wr.Write(p.ParticleID);
                l += wr.Write(p.Mass);
                l += wr.Write(p.X);
                l += wr.Write(p.Y);
                l += wr.Write(p.Z);
                l += wr.Write(p.Vx);
                l += wr.Write(p.Vy);
                l += wr.Write(p.Vz);
                l += wr.Write(p.Phi);
                l += wr.Write(p.Eps);
                return l;
            }

            public CParticle toCParticle()
            {
                return new CParticle(pid,X,Y,Z);
            }

            public Particle toParticle()
            {
                return new Particle(ParticleType.DARK, pid, mass, x, y, z, vx, vy, vz);
            }

            #region IEquatable<BDarkParticle> Members

            public bool Equals(BDarkParticle other)
            {
                return ParticleID == other.ParticleID;
            }

            #endregion

            #region IComparable<Particle> Members

            public int CompareTo(BDarkParticle other)
            {
                return (int)(ParticleID - other.ParticleID);
            }

            #endregion
        }

        public class BStarParticle : IEquatable<BStarParticle>, IComparable<BStarParticle>
        {
            public long pid;
            public float mass;
            public float x;
            public float y;
            public float z;
            public float vx;
            public float vy;
            public float vz;
            public float phi;
            public float metal;
            public float tform;
            public float eps;

            public long ParticleID { get { return pid; } set { pid = value; } }
            public float Mass { get { return mass; } set { mass = value; } }
            public float X { get { return x; } set { x = value; } }
            public float Y { get { return y; } set { y = value; } }
            public float Z { get { return z; } set { z = value; } }
            public float Vx { get { return vx; } set { vx = value; } }
            public float Vy { get { return vy; } set { vy = value; } }
            public float Vz { get { return vz; } set { vz = value; } }
            public float Phi { get { return phi; } set { phi = value; } }
            public float Metal { get { return metal; } set { metal = value; } }
            public float Tform { get { return tform; } set { tform = value; } }
            public float Eps { get { return eps; } set { eps = value; } }

            public static BStarParticle Read(DryadBinaryReader rd)
            {
                BStarParticle p = new BStarParticle();
                p.ParticleID = rd.ReadInt64();
                p.Mass = rd.ReadSingle();
                p.X = rd.ReadSingle();
                p.Y = rd.ReadSingle();
                p.Z = rd.ReadSingle();
                p.Vx = rd.ReadSingle();
                p.Vy = rd.ReadSingle();
                p.Vz = rd.ReadSingle();
                p.Phi = rd.ReadSingle();
                p.Metal = rd.ReadSingle();
                p.Tform = rd.ReadSingle();
                p.Eps = rd.ReadSingle();
                return p;
            }

            public static int Write(DryadBinaryWriter wr, BStarParticle p)
            {
                int l = 0;
                l += wr.Write(p.ParticleID);
                l += wr.Write(p.Mass);
                l += wr.Write(p.X);
                l += wr.Write(p.Y);
                l += wr.Write(p.Z);
                l += wr.Write(p.Vx);
                l += wr.Write(p.Vy);
                l += wr.Write(p.Vz);
                l += wr.Write(p.Phi);
                l += wr.Write(p.Metal);
                l += wr.Write(p.Tform);
                l += wr.Write(p.Eps);
                return l;
            }

            public CParticle toCParticle()
            {
                return new CParticle(pid, X, Y, Z);
            }

            public Particle toParticle()
            {
                return new Particle(ParticleType.STAR, pid, mass, x, y, z, vx, vy, vz);
            }

            #region IEquatable<BGasParticle> Members

            public bool Equals(BStarParticle other)
            {
                return ParticleID == other.ParticleID;
            }

            #endregion

            #region IComparable<Particle> Members

            public int CompareTo(BStarParticle other)
            {
                return (int)(ParticleID - other.ParticleID);
            }

            #endregion
        }


        public class CParticle : IEquatable<CParticle>, IComparable<CParticle>
        {
            long grid;
            long pid;
            long gid;
            float[] coords;
            
            public CParticle(LineRecord l)
                : this(l.line)
            {
            }

            public CParticle(string l)
            {
                string[] str = l.Split(',');

                grid = 0;
                pid = (Int32.Parse(str[0]) << 32)|Int32.Parse(str[1]);
                gid = 0;

                coords = new float[] { 
                    Single.Parse(str[3]),
                    Single.Parse(str[4]),
                    Single.Parse(str[5])
                };
            }

            public CParticle(long id, float x1, float y1, float z1)
            {
                grid = 0;
                pid = id;
                gid = 0;
                coords = new float[] { x1, y1, z1 };
            }

            public CParticle(long grid1, long pid1,long gid1,float x1, float y1, float z1)
            {
                grid = grid1;
                pid = pid1;
                gid = gid1;
                coords = new float[] { x1, y1, z1 };
            }

            /*
            [FieldMapping("p.X", "X")]
            [FieldMapping("p.Y", "Y")]
            [FieldMapping("p.Z", "Z")]
             */
            //public CParticle(Particle p)
            //{
            //    grid = 0;
            //    pid = (p.SnapshotID << 32) | p.ParticleID;
            //    gid = 0;
            //    coords = new float[] { p.Value.X, p.Value.Y, p.Value.Z };
            //}

            //[FieldMapping("g", "GridID")]
            /*
            [FieldMapping("p.ParticleID", "ParticleID")]
            [FieldMapping("p.X", "X")]
            [FieldMapping("p.Y", "Y")]
            [FieldMapping("p.Z", "Z")]
             */
            //public CParticle(long g, Particle p)
            //{
            //    grid = g;
            //    pid = p.ParticleID;
            //    gid = 0;
            //    coords = new float[] { p.Value.X, p.Value.Y, p.Value.Z };
            //}

            //[FieldMapping("gr", "GridID")]
            //[FieldMapping("g", "Group")]
            //[FieldMapping("p.ParticleID", "ParticleID")]
            //[FieldMapping("p.X", "X")]
            //[FieldMapping("p.Y", "Y")]
            //[FieldMapping("p.Z", "Z")]
            public CParticle(long gr, long g, CParticle p)
            {
                grid = gr;
                pid = p.ParticleID;
                gid = g;
                coords = new float[] { p.X, p.Y, p.Z };
            }

            //[FieldMapping("gr", "GridID")]
            //[FieldMapping("p.ParticleID", "ParticleID")]
            //[FieldMapping("p.X", "X")]
            //[FieldMapping("p.Y", "Y")]
            //[FieldMapping("p.Z", "Z")]
            public CParticle(long gr, CParticle p)
            {
                grid = gr;
                pid = p.ParticleID;
                gid = 0;
                coords = new float[] { p.X, p.Y, p.Z };
            }

            public float[] Coordinate
            {
                get { return coords; }
            }

            public long ParticleID { get { return pid; } }
            public long GridID { get { return grid; } set { grid = value; } }

            public float X { get { return coords[0]; } set { coords[0] = value; } }
            public float Y { get { return coords[1]; } set { coords[1] = value; } }
            public float Z { get { return coords[2]; } set { coords[2] = value; } }

            public float[] getLowerBound(float r)
            {
                return new float[3] { X - r, Y - r, Z - r };
            }

            public float[] getUpperBound(float r)
            {
                return new float[3] { X + r, Y + r, Z + r };
            }

            public bool equalCoordinates(float[] o)
            {
                return X == o[0] && Y == o[1] && Z == o[2];
            }
            public bool equalCoordinates(CParticle o)
            {
                return X == o.X && Y == o.Y && Z == o.Z;
            }

            public long Group
            {
                get { return gid; }
                set { gid = value; }
            }

            public void setBounds(float[] low, float[] high, float r)
            {
                low[0] = X - r; high[0] = X + r;
                low[1] = Y - r; high[1] = Y + r;
                low[2] = Z - r; high[2] = Z + r;
            }

            // periodic distance
            public double distance(CParticle o, float period)
            {
                double dx = Math.Min(Math.Abs(X - o.X), period - Math.Abs(X - o.X));
                double dy = Math.Min(Math.Abs(Y - o.Y), period - Math.Abs(Y - o.Y));
                double dz = Math.Min(Math.Abs(Z - o.Z), period - Math.Abs(Z - o.Z));
                return Math.Sqrt(dx * dx + dy * dy + dz * dz);
            }

            public double distance(CParticle o)
            {
                double dx = (X - o.X);
                double dy = (Y - o.Y);
                double dz = (Z - o.Z);
                return Math.Sqrt(dx * dx + dy * dy + dz * dz);
            }

            public void setGroup(int g) { gid = g; }

            public override int GetHashCode() { return (int)pid; }

            #region IEquatable<CParticle> Members

            public bool Equals(CParticle other)
            {
                return ParticleID == other.ParticleID;
            }

            #endregion

            #region IComparable<CParticle> Members

            public int CompareTo(CParticle other)
            {
                if ( this.ParticleID == other.ParticleID ) return 0;
                if (this.ParticleID < other.ParticleID) return -1;
                return 1;
            }

            #endregion
            public override bool Equals(object obj)
            {
                if (obj is CParticle)
                {
                    return this.Equals((CParticle)obj);
                }
                return false;
            }
            public static bool operator ==(CParticle x, CParticle y)
            {
                return x.ParticleID == y.ParticleID;
            }
            public static bool operator !=(CParticle x, CParticle y)
            {
                return x.ParticleID != y.ParticleID;
            }

            public static CParticle Read(DryadBinaryReader rd)
            {
                rd.ReadInt32(); // read dimension
                float x = rd.ReadSingle();
                float y = rd.ReadSingle();
                float z = rd.ReadSingle();
                long gid = rd.ReadInt64();
                long grid = rd.ReadInt64();
                long pid = rd.ReadInt64();
                return new CParticle(grid, pid, gid, x, y, z);
            }

            public static CParticle Read(System.IO.BinaryReader rd)
            {
                int l = rd.ReadInt32(); // read dimension
                float x = rd.ReadSingle();
                float y = rd.ReadSingle();
                float z = rd.ReadSingle();
                long gid = rd.ReadInt64();
                long grid = rd.ReadInt64();
                long pid = rd.ReadInt64();
                return new CParticle(grid, pid, gid, x, y, z);
            }

            public static int Write(DryadBinaryWriter wr, CParticle p)
            {
                int l = 0;
                l += wr.Write(p.Coordinate.Length);
                l += wr.Write(p.X);
                l += wr.Write(p.Y);
                l += wr.Write(p.Z);
                l += wr.Write(p.Group);
                l += wr.Write(p.GridID);
                l += wr.Write(p.ParticleID);
                return l;
            }

            public override string ToString()
            {
                return String.Format("PID={0},GROUP={1},GRID={2},[{3},{4},{5}]", ParticleID, Group, GridID, X, Y, Z);
            }
        }
}
