﻿/*
 * Copyright YongChul Kwon 2009
 * 
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

namespace dFoF.Common
{
    public class KDTree<T>
    {
        public delegate float[] CoordinateExtractor<T>(T v);
        public delegate bool Predicate(T v);

        public class CoordComparer : IComparer<T>
        {
            private CoordinateExtractor<T> m_extractor;
            private int m_d;

            public CoordComparer(CoordinateExtractor<T> ex)
            {
                m_extractor = ex;
                m_d = 0;
            }

            #region IComparer<T> Members

            public int Compare(T x, T y)
            {
                float cx = m_extractor(x)[m_d];
                float cy = m_extractor(y)[m_d];
                return (cx == cy) ? 0 : ((cx < cy) ? -1 : 1);
            }

            public int Dimension { get { return m_d; } set { m_d = value; } }

            #endregion
        }

        private int m_k;
        private Node m_root;
        private Random m_rand;
        private CoordinateExtractor<T> m_extractor;
        private float[] m_low;
        private float[] m_up;

        private CoordComparer m_comp;

        private class Node
        {
            float[] m_coord;
            T       m_value;
            Node m_left;
            Node m_right;

            public Node(float[] c, T val) {
                m_coord = c; m_value = val;
                m_left = null; m_right = null;
            }

            public float[] Coordinate { get { return m_coord; } }
            public T Value { get { return m_value; } set { m_value = value; } }
            public Node Left { get { return m_left; } set { m_left = value; } }
            public Node Right { get { return m_right; } set { m_right = value; } }
            public bool isLeaf { get { return (m_left == null && m_right == null); } }
        }
        
        public KDTree(ref T[] src, int d, CoordinateExtractor<T> e)
        {
            m_k = d;
            m_rand = new Random();
            m_extractor = e;
            m_comp = new CoordComparer(m_extractor);
            m_root = buildTree(ref src, 0, src.Length-1, 0);
            //m_root = buildTree2(ref src, 0, src.Length, 0);
            m_low = new float[m_k];
            m_up = new float[m_k];

        }

        private Node buildTree(ref T[] src, int from, int to, int depth)
        {
            if ( from > to ) {
                return null;
            } else if ( from == to ) {
                float[] c = m_extractor(src[from]);
                return new Node(c,src[from]);
            } else {
                int len = to - from + 1;
                int axis = depth % m_k;
                int med  = len >> 1;
                Util.FindKMedian<T,float>(med+1,ref src,from,to,m_rand,(x) => m_extractor(x)[axis],(x,y) => x <= y);

                float[] c = m_extractor(src[from+med]);

                Node n = new Node(c,src[from+med]);
                n.Left = buildTree(ref src,from, from + med - 1,depth+1);
                n.Right = buildTree(ref src,from + med + 1,to,depth+1);
                return n;
            }
        }

        // build tree using sorting
        private Node buildTree2(ref T[] src, int from, int length, int depth)
        {
            if ( length <= 0 )
            {
                return null;
            }
            else if ( length == 1 )
            {
                float[] c = m_extractor(src[from]);
                return new Node(c, src[from]);
            }
            else
            {
                int endIdx = from + length - 1;
                m_comp.Dimension = depth % m_k;
                Array.Sort(src, from, length, m_comp);

                // median
                int med = from + (length >> 1);
                float[] c = m_extractor(src[med]);

                Node n = new Node(c, src[med]);
                n.Left = buildTree2(ref src, from, med - from, depth + 1);
                n.Right = buildTree2(ref src, med + 1, endIdx - med, depth + 1);
                return n;
            }
        }

        public IEnumerable<T> range(float[] lowk, float[] uppk)
        {
            List<T> b = new List<T>();
            return range(lowk, uppk, (x) => true, b);
        }

        public IEnumerable<T> range(float[] lowk, float[] uppk, List<T> buf)
        {
            return range(lowk, uppk, (x) => true, buf);
        }

        public IEnumerable<T> range(float[] lowk, float[] uppk, Predicate p, List<T> buf)
        {
            rsearch(lowk, uppk, m_root, 0, p, buf);
            return buf;
        }

        private void rsearch(float[] lowk, float[] uppk, Node t, int lev, Predicate p, List<T> buf)
        {
            if (t == null) return;

            float[] c = t.Coordinate;

            if (lowk[lev] <= c[lev])
            {
                rsearch(lowk, uppk, t.Left, (lev + 1) % m_k, p, buf);
            }
            int j;
            for (j = 0; j < m_k && lowk[j] <= c[j] && c[j] <= uppk[j]; j++);
            if (j == m_k)
            {
                if ( p(t.Value) ) buf.Add(t.Value);
            }
            if (c[lev] <= uppk[lev])
            {
                rsearch(lowk, uppk, t.Right, (lev + 1) % m_k, p, buf);
            }
        }

        public IEnumerable<T> range(float[] coord, float d, float period, Predicate p)
        {
            List<T> b = new List<T>();
            return range(0, coord, d, period, p, b);
        }

        public IEnumerable<T> range(T o, float d, float period, Predicate p, List<T> buf)
        {
            return range(0, m_extractor(o), d, period, p, buf);
        }

        public IEnumerable<T> range(int k,float[] coord, float d, float period, Predicate p, List<T> buf)
        {
            if (k == m_k)
            {
                rsearch(m_low, m_up, m_root, 0, p, buf);
            } else {
                // compute bound
                float cmin = -0.5f * period;
                float cmax = 0.5f * period;
                float lb = cmin + d;
                float ub = cmax - d;

                float c = coord[k];
                
                if (lb < c && c < ub)
                {
                    m_low[k] = c - d;
                    m_up[k] = c + d;
                    range(k + 1, coord, d, period, p, buf);
                }
                else if (c <= lb)
                {
                    m_low[k] = cmin;
                    m_up[k] = c + d;
                    range(k + 1, coord, d, period, p, buf);

                    m_low[k] = c - d + period;
                    m_up[k] = cmax;
                    range(k + 1, coord, d, period, p, buf);
                }
                else
                {
                    m_low[k] = c - d;
                    m_up[k] = cmax;
                    range(k + 1, coord, d, period, p, buf);

                    m_low[k] = cmin;
                    m_up[k] = c + d - period;
                    range(k + 1, coord, d, period, p, buf);

                }
            }
            return buf;
        }

        private string getLabel(int d, float[] coords)
        {
            StringBuilder buf = new StringBuilder();
            for (int i = 0; i < coords.Length; ++i)
            {
                if (i == d)
                {
                    buf.Append("<font color=\"red\">").Append(String.Format("{0:0.0000}",coords[i])).Append("</font>");
                }
                else
                {
                    buf.Append(String.Format("{0:0.0000}", coords[i]));
                }
                buf.Append(',');
            }
            buf.Length = buf.Length - 1;
            return buf.ToString();
        }

        private void dump(StreamWriter wr, Node n,int d)
        {
            if (n == null) return;
            wr.WriteLine("{0} [shape=box,label=<{0}<br/>{1}>];", n.Value.GetHashCode(), getLabel(d,n.Coordinate));
            dump(wr, n.Left, (d+1) % m_k);
            dump(wr, n.Right, (d+1) % m_k);
            if (n.Left != null) wr.WriteLine("{0} -> {1} [label=L];", n.Value.GetHashCode(), n.Left.Value.GetHashCode());
            if (n.Right != null) wr.WriteLine("{0} -> {1} [label=R];", n.Value.GetHashCode(), n.Right.Value.GetHashCode());
        }

        public void toDot(string fn)
        {
            using (StreamWriter wr = new StreamWriter(fn))
            {
                wr.WriteLine("digraph kdtree {");
                dump(wr, m_root, 0);
                wr.WriteLine("}");
                // write up vertexes
            }
        }
    }

    public class KDTreeOPT<T>
    {
        public delegate float[] CoordinateExtractor<T>(T v);
        public delegate bool Predicate(T v);
        public delegate void Visit(T v);

        private static void DefaultVisit(T v) { } // do nothing

        public class CoordComparer : IComparer<T>
        {
            private CoordinateExtractor<T> m_extractor;
            private int m_d;

            public CoordComparer(CoordinateExtractor<T> ex)
            {
                m_extractor = ex;
                m_d = 0;
            }

            #region IComparer<T> Members

            public int Compare(T x, T y)
            {
                float cx = m_extractor(x)[m_d];
                float cy = m_extractor(y)[m_d];
                return (cx == cy) ? 0 : ((cx < cy) ? -1 : 1);
            }

            public int Dimension { get { return m_d; } set { m_d = value; } }

            #endregion
        }

        private int m_k;
        private Node m_root;
        private Random m_rand;
        private CoordinateExtractor<T> m_extractor;
        private float[] m_low;
        private float[] m_up;

        private CoordComparer m_comp;
        private Predicate m_isRevisit;
        private Visit m_visit;

        public Predicate IsRevisit { get { return m_isRevisit; } set { m_isRevisit = value; } }
        public Visit DoVisit { get { return m_visit; } set { m_visit = value; } }

        private sealed class Node
        {
            float[] m_coord;
            T m_value;
            Node m_left;
            Node m_right;
            bool m_visited;

            public Node(float[] c, T val)
            {
                m_coord = c; m_value = val;
                m_left = null; m_right = null;
                m_visited = false;
            }

            public float[] Coordinate { get { return m_coord; } }
            public T Value { get { return m_value; } set { m_value = value; } }
            public Node Left { get { return m_left; } set { m_left = value; } }
            public Node Right { get { return m_right; } set { m_right = value; } }
            public bool isLeaf { get { return (m_left == null && m_right == null); } }
            public bool VisitedAll { get { return m_visited; } set { m_visited = value; } }
        }

        public KDTreeOPT(ref T[] src, int d, CoordinateExtractor<T> e)
        {
            m_k = d;
            m_rand = new Random();
            m_extractor = e;
            m_comp = new CoordComparer(m_extractor);
            m_root = buildTree(ref src, 0, src.Length - 1, 0);
            //m_root = buildTree2(ref src, 0, src.Length, 0);
            m_low = new float[m_k];
            m_up = new float[m_k];

            m_isRevisit = (x) => false; // always revisit
            m_visit = DefaultVisit;
        }

        private Node buildTree(ref T[] src, int from, int to, int depth)
        {
            if (from > to)
            {
                return null;
            }
            else if (from == to)
            {
                float[] c = m_extractor(src[from]);
                return new Node(c, src[from]);
            }
            else
            {
                int len = to - from + 1;
                int axis = depth % m_k;
                int med = len >> 1;
                Util.FindKMedian<T, float>(med + 1, ref src, from, to, m_rand, (x) => m_extractor(x)[axis], (x, y) => x <= y);

                float[] c = m_extractor(src[from + med]);

                Node n = new Node(c, src[from + med]);
                n.Left = buildTree(ref src, from, from + med - 1, depth + 1);
                n.Right = buildTree(ref src, from + med + 1, to, depth + 1);

                return n;
            }
        }

        // build tree using sorting
        private Node buildTree2(ref T[] src, int from, int length, int depth)
        {
            if (length <= 0)
            {
                return null;
            }
            else if (length == 1)
            {
                float[] c = m_extractor(src[from]);
                return new Node(c, src[from]);
            }
            else
            {
                int endIdx = from + length - 1;
                m_comp.Dimension = depth % m_k;
                Array.Sort(src, from, length, m_comp);

                // median
                int med = from + (length >> 1);
                float[] c = m_extractor(src[med]);


                Node left = buildTree2(ref src, from, med - from, depth + 1);
                Node right = buildTree2(ref src, med + 1, endIdx - med, depth + 1);
                Node n = new Node(c, src[med]);
                n.Left = left;
                n.Right = right;

                return n;
            }
        }

        public IEnumerable<T> range(float[] lowk, float[] uppk)
        {
            List<T> b = new List<T>();
            return range(lowk, uppk, (x) => true, b);
        }

        public IEnumerable<T> range(float[] lowk, float[] uppk, List<T> buf)
        {
            return range(lowk, uppk, (x) => true, buf);
        }

        public IEnumerable<T> range(float[] lowk, float[] uppk, Predicate p, List<T> buf)
        {
            rsearch(lowk, uppk, m_root, 0, p, buf);
            return buf;
        }

        private void rsearch(float[] lowk, float[] uppk, Node t, int lev, Predicate p, List<T> buf)
        {
            if (t == null) return;

            // decide whether to traverse subtree or not.
            if (t.VisitedAll) return;

            int prevSize = buf.Count;
            float[] c = t.Coordinate;

            if (lowk[lev] <= c[lev])
            {
                rsearch(lowk, uppk, t.Left, (lev + 1) % m_k, p, buf);
            }
            if (c[lev] <= uppk[lev])
            {
                rsearch(lowk, uppk, t.Right, (lev + 1) % m_k, p, buf);
            }

            int j;
            for (j = 0; j < m_k && lowk[j] <= c[j] && c[j] <= uppk[j]; j++) ;
            if (j == m_k)
            {
                if (!m_isRevisit(t.Value) && p(t.Value))
                {
                    m_visit(t.Value);
                    buf.Add(t.Value);
                }
            }

            // was there any change?
            if ( prevSize < buf.Count ) {
                // check whether we can prune this subtree for future
                bool l = (t.Left == null) ? true : t.Left.VisitedAll;
                bool r = (t.Right == null) ? true : t.Right.VisitedAll;
                t.VisitedAll = l && r && m_isRevisit(t.Value);
            }
        }

        public IEnumerable<T> range(float[] coord, float d, float period, Predicate p)
        {
            List<T> b = new List<T>();
            return range(0, coord, d, period, p, b);
        }

        public IEnumerable<T> range(T o, float d, float period, Predicate p, List<T> buf)
        {
            return range(0, m_extractor(o), d, period, p, buf);
        }

        public IEnumerable<T> range(int k, float[] coord, float d, float period, Predicate p, List<T> buf)
        {
            if (k == m_k)
            {
                rsearch(m_low, m_up, m_root, 0, p, buf);
            }
            else
            {
                // compute bound
                float cmin = -0.5f * period;
                float cmax = 0.5f * period;
                float lb = cmin + d;
                float ub = cmax - d;

                float c = coord[k];

                if (lb < c && c < ub)
                {
                    m_low[k] = c - d;
                    m_up[k] = c + d;
                    range(k + 1, coord, d, period, p, buf);
                }
                else if (c <= lb)
                {
                    m_low[k] = cmin;
                    m_up[k] = c + d;
                    range(k + 1, coord, d, period, p, buf);

                    m_low[k] = c - d + period;
                    m_up[k] = cmax;
                    range(k + 1, coord, d, period, p, buf);
                }
                else
                {
                    m_low[k] = c - d;
                    m_up[k] = cmax;
                    range(k + 1, coord, d, period, p, buf);

                    m_low[k] = cmin;
                    m_up[k] = c + d - period;
                    range(k + 1, coord, d, period, p, buf);

                }
            }
            return buf;
        }

        private string getLabel(int d, float[] coords)
        {
            StringBuilder buf = new StringBuilder();
            for (int i = 0; i < coords.Length; ++i)
            {
                if (i == d)
                {
                    buf.Append("<font color=\"red\">").Append(String.Format("{0:0.0000}", coords[i])).Append("</font>");
                }
                else
                {
                    buf.Append(String.Format("{0:0.0000}", coords[i]));
                }
                buf.Append(',');
            }
            buf.Length = buf.Length - 1;
            return buf.ToString();
        }

        private void dump(StreamWriter wr, Node n, int d)
        {
            if (n == null) return;
            wr.WriteLine("{0} [shape=box,label=<{0}<br/>{1}>];", n.Value.GetHashCode(), getLabel(d, n.Coordinate));
            dump(wr, n.Left, (d + 1) % m_k);
            dump(wr, n.Right, (d + 1) % m_k);
            if (n.Left != null) wr.WriteLine("{0} -> {1} [label=L];", n.Value.GetHashCode(), n.Left.Value.GetHashCode());
            if (n.Right != null) wr.WriteLine("{0} -> {1} [label=R];", n.Value.GetHashCode(), n.Right.Value.GetHashCode());
        }

        public void toDot(string fn)
        {
            using (StreamWriter wr = new StreamWriter(fn))
            {
                wr.WriteLine("digraph kdtree {");
                dump(wr, m_root, 0);
                wr.WriteLine("}");
                // write up vertexes
            }
        }
    }
}
