﻿/*
 * Copyright YongChul Kwon 2009
 * 
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Text;
using System.Threading;

using LinqToDryad;
using dFoF.Common;
using System.IO;
using System.IO.Compression;

namespace dFoF
{   
    // grid function. grid id. 64bit = depth 21.

    // Grid order
    //  2 5
    // 1 6
    //  3 4
    // 0 7

    [Serializable]
    public class Grid : IEquatable<Grid>, IComparable<Grid>
    {
        long gid;

        // low point
        float m_x;
        float m_y;
        float m_z;

        // high point
        float m_X;
        float m_Y;
        float m_Z;

        // split point
        float m_sx;
        float[] m_sy;
        float[] m_sz;

        public Grid(long id, float x1, float y1, float z1, float x2, float y2, float z2)
        {
            gid = id;
            m_x = x1; m_y = y1; m_z = z1;
            m_X = x2; m_Y = y2; m_Z = z2;
            m_sx = float.NaN;
            m_sy = null;
            m_sz = null;
        }

        public Grid[] Split(float x, float[] y, float[] z)
        {
            Grid[] subGrids = new Grid[8];
            m_sx = x;
            m_sy = y.Clone() as float[];
            m_sz = z.Clone() as float[];

            subGrids[0] = new Grid((gid << 3) | 0, m_x, m_y, m_z, x, y[0], z[0]);
            subGrids[1] = new Grid((gid << 3) | 1, m_x, m_y, z[0], x, y[0], m_Z);
            subGrids[2] = new Grid((gid << 3) | 2, m_x, y[0], m_z, x, m_Y, z[1]);
            subGrids[3] = new Grid((gid << 3) | 3, m_x, y[0], z[1], x, m_Y, m_Z);

            subGrids[4] = new Grid((gid << 3) | 4, x, m_y, m_z, m_X, y[1], z[2]);
            subGrids[5] = new Grid((gid << 3) | 5, x, m_y, z[2], m_X, y[1], m_Z);
            subGrids[6] = new Grid((gid << 3) | 6, x, y[1], m_z, m_X, m_Y, z[3]);
            subGrids[7] = new Grid((gid << 3) | 7, x, y[1], z[3], m_X, m_Y, m_Z);

            return subGrids;
        }

        public Grid[] Split()
        {
            m_sx = (m_x + m_X) * 0.5f;
            m_sy = new float[2] { (m_y + m_Y) * 0.5f, (m_y + m_Y) * 0.5f };
            m_sz = new float[4] { (m_z + m_Z) * 0.5f, (m_z + m_Z) * 0.5f, (m_z + m_Z) * 0.5f, (m_z + m_Z) * 0.5f };

            Grid[] subGrids = new Grid[8];
            
            subGrids[0] = new Grid((gid << 3) | 0, m_x, m_y, m_z, m_sx, m_sy[0], m_sz[0]);
            subGrids[1] = new Grid((gid << 3) | 1, m_x, m_y, m_sz[0], m_sx, m_sy[0], m_Z);
            subGrids[2] = new Grid((gid << 3) | 2, m_x, m_sy[0], m_z, m_sx, m_Y, m_sz[1]);
            subGrids[3] = new Grid((gid << 3) | 3, m_x, m_sy[0], m_sz[1], m_sx, m_Y, m_Z);

            subGrids[4] = new Grid((gid << 3) | 4, m_sx, m_y, m_z, m_X, m_sy[1], m_sz[2]);
            subGrids[5] = new Grid((gid << 3) | 5, m_sx, m_y, m_sz[2], m_X, m_sy[1], m_Z);
            subGrids[6] = new Grid((gid << 3) | 6, m_sx, m_sy[1], m_z, m_X, m_Y, m_sz[3]);
            subGrids[7] = new Grid((gid << 3) | 7, m_sx, m_sy[1], m_sz[3], m_X, m_Y, m_Z);

            return subGrids;
        }

        public bool isCore(float x, float y, float z, float R)
        {
            // check whether it's within core range
            return (m_x + R < x && x < m_X - R) && (m_y + R < y && y < m_Y - R) && (m_z + R < z && z < m_Z - R);
        }

        public bool isBoundary(float x, float y, float z, float R)
        {
            return !isCore(x, y, z, R);
        }

        public bool sanityCheck(float R)
        {
            if (m_x >= m_X) return false;
            if (m_y >= m_Y) return false;
            if (m_z >= m_Z) return false;
            if (m_X - m_x <= 2.0*R) return false;
            if (m_Y - m_y <= 2.0 * R) return false;
            if (m_Z - m_z <= 2.0 * R) return false;
            return true;
        }

        public long ID { get { return gid; } }
        public long ParentID { get { return gid >> 3; } }

        public int GetGridID(float x, float y, float z)
        {
            int id = 0;

            if (x < m_sx)
            {
                // 0 1 2 3
                if (y < m_sy[0])
                {
                    id = (z < m_sz[0]) ? 0 : 1;
                }
                else
                {
                    id = (z < m_sz[1]) ? 2 : 3;
                }
            }
            else
            {
                // 4 5 6 7
                if (y < m_sy[1])
                {
                    id = (z < m_sz[2]) ? 4 : 5;
                }
                else
                {
                    id = (z < m_sz[3]) ? 6 : 7;
                }
            }
            return id;
        }

        #region IEquatable<Grid> Members

        bool IEquatable<Grid>.Equals(Grid other)
        {
            return gid == other.gid 
                    && m_x == other.m_x && m_y == other.m_y && m_z == other.m_z
                    && m_X == other.m_X && m_Y == other.m_Y && m_Z == other.m_Z;
        }

        #endregion

        #region IComparable<Grid> Members

        int IComparable<Grid>.CompareTo(Grid other)
        {
            return (int)(gid - other.gid);
        }

        #endregion

        private float randomPt(bool inside,float low, float high, float R,Random rand)
        {
            if (inside)
            {
                return (float)(low + R + (high - low - 2.0f * R) * rand.NextDouble());
            }
            else
            {
                double xx = rand.NextDouble() * 2.0 * R;
                return (float)((xx < R) ? low + xx : xx + high - 2.0 * R);
            }
        }

        public CParticle randomPoint(Random rand,int pid, float R, bool isCore)
        {
            float x;
            float y;
            float z;

            int mask = rand.Next();

            x = randomPt((mask & 0x01) == 0, m_x, m_X, R, rand);
            y = randomPt((mask & 0x02) == 0, m_y, m_Y, R, rand);
            z = randomPt((mask & 0x04) == 0, m_z, m_Z, R, rand);

            return new CParticle(pid, x, y, z);
        }

        public override string ToString()
        {
            if (m_sy == null)
            {
                return String.Format("ID={0},LOW=[{1},{2},{3}],HIGH=[{4},{5},{6}]", gid, m_x, m_y, m_z, m_X, m_Y, m_Z);
            }
            else
            {
                return String.Format("ID={0},LOW=[{1},{2},{3}],HIGH=[{4},{5},{6}],SPLIT X={7},SPLIT Y=[{8},{9}],SPLIT Z=[{10},{11},{12},{13}]",gid, m_x, m_y, m_z, m_X, m_Y, m_Z, m_sx, m_sy[0], m_sy[1], m_sz[0], m_sz[1], m_sz[2], m_sz[3]);
            }
        }
    }

    #region Grid tree structure

    [Serializable]
    public class GridTree
    {
        [Serializable]
        public class GridTreeNode
        {
            Grid current;
            GridTreeNode[] children;

            public GridTreeNode(Grid n)
            {
                current = n;
                children = null;
            }

            public GridTreeNode[] Split(float x,float[] y,float[] z) {
                Grid[] subGrids = current.Split(x,y,z);
                children = new GridTreeNode[8];
                for ( int i = 0; i < 8; ++i )
                    children[i] = new GridTreeNode(subGrids[i]);
                return children;
            }

            public void Split(int depth,List<Dictionary<long,Grid>> gridIdx)
            {
                gridIdx[depth].Add(current.ID, current);

                if (depth == 0) return;

                Grid[] specs = current.Split();
                children = new GridTreeNode[8];
                for (int i = 0; i < children.Length; ++i)
                {
                    children[i] = new GridTreeNode(specs[i]);
                    children[i].Split(depth - 1,gridIdx);
                }
            }

            public bool isLeaf { get { return children == null; } }
            public Grid GridSpec { get { return current; } }
            public GridTreeNode this[int index]
            {
                get
                {
                    return children[index];
                }
            }
            public GridTreeNode[] Children { get { return children; } set { children = value; } }
        }

//        [NonSerialized]
        private GridTreeNode m_root;
        private int m_leaves;
        private List<Dictionary<long, Grid>> m_grids;
        private float m_R;

        public int Count { get { return m_leaves; } }

            /*
        public GridTree(int depth, float unit, float R)
        {
            // instantiate grid tree
            m_root = new GridTreeNode(new Grid(0, unit, 0.0f, 0.0f, 0.0f));
            m_grids = new List<Dictionary<long, Grid>>();
            m_R = R;
            for (int i = 0; i <= depth; ++i)
            {
                m_grids.Add(new Dictionary<long, Grid>());
            }
            m_root.Split(depth,m_grids);

        }
             */
        public GridTree(GridTree.GridTreeNode root,List<Dictionary<long,Grid>> grids,int depth,float R) {
            m_root = root;
            m_R = R;
            m_grids = grids;
        }

        public long[] toBuckets(int lv)
        {
            //long[] r = m_grids[lv].Keys.ToArray();
            long[] r = m_grids[lv].Keys.Select((g) => g << (lv * 3)).ToArray();
            Array.Sort(r);
            long[] bucket = new long[r.Length - 1];
            Array.Copy(r, 1, bucket, 0, bucket.Length);
            for (int i = 0; i < bucket.Length; ++i)
            {
                bucket[i] -= 1;
            }
            //Array.Resize(ref r,r.Length - 1);
            return bucket;
        }

        public bool isBoundary(CParticle p, int lv)
        {
            long gridId = p.GridID >> (lv * 3);
            Grid g = m_grids[lv][gridId];
            return g.isBoundary(p.X, p.Y, p.Z, m_R);
        }

        public bool isBoundary(CParticle p, int[] lv)
        {
            long gridId = p.GridID >> (lv[0] * 3);
            Grid g = m_grids[lv[0]][gridId];
            return g.isBoundary(p.X, p.Y, p.Z, m_R);
        }

        public long GetGridID(float x, float y, float z)
        {
            return GetGrid(x, y, z).ID;
        }

        public long GetGridID(CParticle p)
        {
            return GetGridID(p.X, p.Y, p.Z);
        }

        public Grid GetGrid(float x, float y, float z)
        {
            GridTreeNode node = m_root;
            while (!node.isLeaf)
            {
                int gid = node.GridSpec.GetGridID(x, y, z);
                node = node[gid];
            }
            return node.GridSpec;
        }

        public IEnumerable<Grid> AsEnumerable()
        {
            // traverse tree and add all leaf nodes
            Stack<GridTreeNode> stack = new Stack<GridTreeNode>();
            List<Grid> results = new List<Grid>();

            stack.Push(m_root);

            while (stack.Count() > 0)
            {
                GridTreeNode node = stack.Pop();
                if (node.isLeaf)
                {
                    // add it to the list
                    results.Add(node.GridSpec);
                }
                else
                {
                    // expand
                    foreach (GridTreeNode child in node.Children)
                    {
                        stack.Push(child);
                    }
                }
            }

            m_leaves = results.Count;

            return results.OrderBy((g) => g.ID);
        }

        public IQueryable<Grid> ToDryadTable()
        {
            return AsEnumerable().ToDryadTable();
        }

        public IQueryable<CParticle> GenerateTestData(int n)
        {
            Random rand = new Random(0);
            int pid = 0;

            // traverse tree and add all leaf nodes
            Stack<GridTreeNode> stack = new Stack<GridTreeNode>();
            List<CParticle> results = new List<CParticle>();

            stack.Push(m_root);

            while (stack.Count() > 0)
            {
                GridTreeNode node = stack.Pop();
                if (node.isLeaf)
                {
                    // add it to the list
                    for (int i = 0; i < n; ++i)
                    {
                        results.Add(node.GridSpec.randomPoint(rand, ++pid, m_R, true));
                        results.Add(node.GridSpec.randomPoint(rand, ++pid, m_R, false));
                    }
                }
                else
                {
                    // expand
                    foreach (GridTreeNode child in node.Children)
                    {
                        stack.Push(child);
                    }
                }
            }
            return results.ToDryadTable();
        }

        public void printTree(TextWriter w,float R)
        {
            for ( int i = 0; i < m_grids.Count; ++i ) {
                Dictionary<long,Grid> d = m_grids[i];
                foreach (Grid grid in d.Values.OrderBy((g) => g.ID))
                {
                    Debug.Assert(grid.sanityCheck(R));
                    w.WriteLine("Depth={0},{1}",i,grid);
                }
            }
        }
    }

    #endregion

    public class Program
    {

        #region friend of friend

        [Homomorphic]
        [Resource(IsStateful = true)]
        public static IEnumerable<CParticle> fof_naive(IEnumerable<CParticle> particles)
        {
            CParticle[] l = particles.ToArray();

            Console.WriteLine("{0}: Loading done. {1} particles", DateTime.Now, l.Length);
            KDTree<CParticle> tree = new KDTree<CParticle>(ref l, 3, (p) => p.Coordinate);

            Stack<CParticle> stack = new Stack<CParticle>();
            List<CParticle> buf = new List<CParticle>();
            float[] lowk = new float[3];
            float[] highk = new float[3];

            int i;
            int g = 0;
            int numParticles = 0;
            int numGroups = 0;

            for (i = 0; i < l.Length; ++i)
            {
                CParticle seed = l[i];
                if (seed.Group > 0) continue;

                long gid = seed.ParticleID + 1;

                // recursively search
                stack.Push(seed);
                int groupSize = 1; // counting seed

                while (stack.Count > 0)
                {
                    CParticle p = stack.Pop();
                    if (p.Group > 0) continue;
                    p.Group = gid; // mark visited

                    p.setBounds(lowk, highk, Radius);
                    buf.Clear();
                    tree.range(lowk, highk, (x) => x.Group == 0 && p.distance(x) <= Radius, buf);
                    foreach (CParticle x in buf)
                    {
                        x.Group = -1;
                        stack.Push(x);
                    }
                    groupSize += buf.Count;
                }

                /*
                if (groupSize > 1000)
                {
                    Console.WriteLine("{0}: |group({1})| = {2}", DateTime.Now, seed.Group, groupSize);
                }
                 * */

                if (groupSize >= MinParticles)
                {
                    ++numGroups;
                    numParticles += groupSize;
                }
                ++g;
            }

            Console.WriteLine("{0}: {1} processed. found {2} groups. {3} particles in {4} groups. Done with FOF", DateTime.Now, i, g, numParticles, numGroups);

            return l;
        }

        [Homomorphic]
        [Resource(IsStateful = true)]
        public static IEnumerable<CParticle> fof_opt(IEnumerable<CParticle> particles)
        {
            CParticle[] l = particles.ToArray();
            //CParticle[] l = particles;

            Console.WriteLine("{0}: Loading done. {1} particles", DateTime.Now, l.Length);
            KDTreeOPT<CParticle> tree = new KDTreeOPT<CParticle>(ref l, 3, (p) => p.Coordinate);

            tree.IsRevisit = (p) => p.Group != 0;
            tree.DoVisit = (p) => { p.Group = -1; };

            Stack<CParticle> stack = new Stack<CParticle>();
            List<CParticle> buf = new List<CParticle>();
            float[] lowk = new float[3];
            float[] highk = new float[3];

            int i;
            int g = 0;
            int numParticles = 0;
            int numGroups = 0;

            for (i = 0; i < l.Length; ++i)
            {
                CParticle seed = l[i];
                if (seed.Group > 0) continue;

                long gid = seed.ParticleID + 1;

                // recursively search
                stack.Push(seed);
                int groupSize = 1; // counting seed

                while ( stack.Count > 0)
                {
                    CParticle p = stack.Pop();
                    if (p.Group > 0) continue;
                    p.Group = gid; // mark visited

                    p.setBounds(lowk, highk, Radius);
                    buf.Clear();
                    tree.range(lowk, highk, (x) => p.distance(x) <= Radius, buf);
                    foreach (CParticle x in buf)
                    {
                        stack.Push(x);
                    }
                    groupSize += buf.Count;
                }

                /*
                if (groupSize > 1000)
                {
                    Console.WriteLine("{0}: |group({1})| = {2}", DateTime.Now, seed.Group, groupSize);
                }
                 * */

                if (groupSize >= MinParticles)
                {
                    ++numGroups;
                    numParticles += groupSize;
                }
                ++g;
            }

            Console.WriteLine("{0}: {1} processed. found {2} groups. {3} particles in {4} groups. Done with FOF", DateTime.Now, i, g, numParticles, numGroups);

            return l;
        }

        #endregion

        #region link groups across boundary

        [Serializable]
        struct LongDescendingComparer : IComparer<long>
        {

            #region IComparer<long> Members

            public int Compare(long x, long y)
            {
                return (int)(y - x);
            }

            #endregion
        }

        [Homomorphic]
        [Resource(IsStateful = true)]
        public static IEnumerable<Pair<long, long>> ResolveFoF_naive(IEnumerable<CParticle> particles)
        {
            Console.WriteLine("{0}: load begin...", DateTime.Now);
            CParticle[] px = particles.ToArray(); // loaded in memory

            CParticle[][] groups = px.GroupBy((p) => p.Group).Select((g) => g.ToArray()).OrderBy((g) => -1 * g.Length).ToArray();

            Console.WriteLine("{0}: Loading done. {1} particles", DateTime.Now, px.Length);
            KDTree<CParticle> spIndex = new KDTree<CParticle>(ref px, 3, (p) => p.Coordinate);

            int k = 0;

            for (int i = 0; i < px.Length; ++i)
            {
                px[i].GridID = 0;
            }

            Dictionary<long, long> mappings = new Dictionary<long, long>();

            Stack<CParticle> stack = new Stack<CParticle>();
            List<CParticle> buf = new List<CParticle>();
            float[] lowk = new float[3];
            float[] highk = new float[3];

            HashSet<long> gids = new HashSet<long>();
            HashSet<long> mappedGids = new HashSet<long>();

            for (int i = 0; i < groups.Length; ++i)
            {
                CParticle[] group = groups[i];
                gids.Clear();
                gids.Add(group[0].Group);
                for (int j = 0; j < group.Length; ++j)
                {
                    CParticle seed = group[j];
                    if (seed.GridID > 0) continue;

                    // recursively search
                    stack.Push(seed);

                    int groupSize = 1; // counting seed

                    while (stack.Count > 0)
                    {
                        CParticle p = stack.Pop();
                        if (p.GridID > 0) continue;
                        p.GridID = 1;// mark visited

                        p.setBounds(lowk, highk, Radius);
                        buf.Clear();
                        spIndex.range(lowk, highk, (x) => x.GridID == 0 && p.distance(x) <= Radius, buf);

                        foreach (CParticle x in buf)
                        {
                            x.GridID = -1;
                            stack.Push(x);
                            gids.Add(x.Group);
                        }

                        groupSize += buf.Count;
                    }
                }

                // now we got all neighboring groups
                // find all mapped groups

                if (gids.Count > 1)
                {
                    mappedGids.Clear();
                    foreach (long g in gids)
                    {
                        mappedGids.Add(g);
                        if (mappings.ContainsKey(g))
                        {
                            long ngid = g;
                            do
                            {
                                ngid = mappings[ngid];
                                mappedGids.Add(ngid);
                            } while (mappings.ContainsKey(ngid));
                        }
                    }
                    long newGid = mappedGids.Min();
                    foreach (long g in mappedGids)
                    {
                        if (g > newGid)
                            mappings[g] = newGid;
                    }
                } // if all of them are in a single groups, no need to store mapping
            }

            Console.WriteLine("{0}: first phase done", DateTime.Now);

            long[] allGids = mappings.Keys.ToArray();
            Stack<long> path = new Stack<long>();
            foreach (long g in allGids)
            {
                long ngid = mappings[g];
                while ((mappings.ContainsKey(ngid)))
                {
                    path.Push(ngid);
                    ngid = mappings[ngid];
                }
                while (path.Count > 0)
                { // remap for future
                    mappings[path.Pop()] = ngid;
                }
                ++k;
                yield return new Pair<long, long>(g, ngid);
            }

            Console.WriteLine("{0}: Total {1} new mappings were identified", DateTime.Now, k);
        }


        [Homomorphic]
        [Resource(IsStateful = true)]
        public static IEnumerable<Pair<long, long>> ResolveFoF_opt(IEnumerable<CParticle> particles)
        {
            Console.WriteLine("{0}: load begin...", DateTime.Now);
            CParticle[] px = particles.ToArray(); // loaded in memory

            CParticle[][] groups = px.GroupBy((p) => p.Group).Select((g) => g.ToArray()).OrderBy((g) => -1 * g.Length).ToArray();

            Console.WriteLine("{0}: Loading done. {1} particles", DateTime.Now, px.Length);
            KDTreeOPT<CParticle> spIndex = new KDTreeOPT<CParticle>(ref px, 3, (p) => p.Coordinate);

            spIndex.IsRevisit = (p) => p.GridID != 0;
            spIndex.DoVisit = (p) => { p.GridID = -1; };

            int k = 0;

            for (int i = 0; i < px.Length; ++i)
            {
                px[i].GridID = 0;
            }

            Dictionary<long, long> mappings = new Dictionary<long, long>();

            Stack<CParticle> stack = new Stack<CParticle>();
            List<CParticle> buf = new List<CParticle>();
            float[] lowk = new float[3];
            float[] highk = new float[3];

            HashSet<long> gids = new HashSet<long>();
            HashSet<long> mappedGids = new HashSet<long>();

            for ( int i = 0; i < groups.Length; ++i ) {
                CParticle[] group = groups[i];
                gids.Clear();
                gids.Add(group[0].Group);
                for ( int j = 0; j < group.Length; ++j ) {
                    CParticle seed = group[j];
                    if (seed.GridID > 0) continue;

                    // recursively search
                    stack.Push(seed);

                    int groupSize = 1; // counting seed

                    while (stack.Count > 0)
                    {
                        CParticle p = stack.Pop();
                        if (p.GridID > 0) continue;
                        p.GridID = 1;// mark visited

                        p.setBounds(lowk, highk, Radius);
                        buf.Clear();
                        spIndex.range(lowk, highk, (x) => p.distance(x) <= Radius, buf);

                        foreach (CParticle x in buf)
                        {
                            stack.Push(x);
                            gids.Add(x.Group);
                        }

                        groupSize += buf.Count;
                    }
                }

                // now we got all neighboring groups
                // find all mapped groups

                if (gids.Count > 1)
                {
                    mappedGids.Clear();
                    foreach (long g in gids)
                    {
                        mappedGids.Add(g);
                        if (mappings.ContainsKey(g))
                        {
                            long ngid = g;   
                            do
                            {
                                ngid = mappings[ngid];
                                mappedGids.Add(ngid);
                            } while (mappings.ContainsKey(ngid));
                        }
                    }
                    long newGid = mappedGids.Min();
                    foreach (long g in mappedGids)
                    {
                        if ( g > newGid )
                            mappings[g] = newGid;
                    }
                } // if all of them are in a single groups, no need to store mapping
            }

            Console.WriteLine("{0}: first phase done", DateTime.Now);

            long[] allGids = mappings.Keys.ToArray();
            Stack<long> path = new Stack<long>();
            foreach (long g in allGids)
            {
                long ngid = mappings[g];
                while ((mappings.ContainsKey(ngid)))
                {
                    path.Push(ngid);
                    ngid = mappings[ngid];
                }
                while (path.Count > 0)
                { // remap for future
                    mappings[path.Pop()] = ngid;
                }
                ++k;
                yield return new Pair<long, long>(g, ngid);
            }

            Console.WriteLine("{0}: Total {1} new mappings were identified", DateTime.Now, k);
        }


        #endregion

        #region boundary condition
        [Resource(IsStateful = true)]
        public static IEnumerable<Pair<long, long>> ResolveFoF_Final_naive(IEnumerable<CParticle> particles)
        {
            Console.WriteLine("{0}: load begin...", DateTime.Now);
            CParticle[] px = particles.ToArray(); // loaded in memory

            CParticle[][] groups = px.GroupBy((p) => p.Group).Select((g) => g.ToArray()).OrderBy((g) => -1 * g.Length).ToArray();

            Console.WriteLine("{0}: Loading done. {1} particles", DateTime.Now, px.Length);
            KDTree<CParticle> spIndex = new KDTree<CParticle>(ref px, 3, (p) => p.Coordinate);

            int k = 0;

            for (int i = 0; i < px.Length; ++i)
            {
                px[i].GridID = 0;
            }

            Dictionary<long, long> mappings = new Dictionary<long, long>();

            Stack<CParticle> stack = new Stack<CParticle>();
            List<CParticle> buf = new List<CParticle>();
            float[] lowk = new float[3];
            float[] highk = new float[3];

            HashSet<long> gids = new HashSet<long>();
            HashSet<long> mappedGids = new HashSet<long>();

            for (int i = 0; i < groups.Length; ++i)
            {
                CParticle[] group = groups[i];
                gids.Clear();
                gids.Add(group[0].Group);
                for (int j = 0; j < group.Length; ++j)
                {
                    CParticle seed = group[j];
                    if (seed.GridID > 0) continue;

                    // recursively search
                    stack.Push(seed);

                    int groupSize = 1; // counting seed

                    while (stack.Count > 0)
                    {
                        CParticle p = stack.Pop();
                        if (p.GridID > 0) continue;
                        p.GridID = 1;// mark visited

                        buf.Clear();
                        spIndex.range(p, Radius, Period, (x) => x.GridID == 0 && p.distance(x, Period) <= Radius, buf);

                        foreach (CParticle x in buf)
                        {
                            x.GridID = -1;
                            stack.Push(x);
                            gids.Add(x.Group);
                        }

                        groupSize += buf.Count;
                    }
                }

                // now we got all neighboring groups
                // find all mapped groups

                if (gids.Count > 1)
                {
                    mappedGids.Clear();
                    foreach (long g in gids)
                    {
                        mappedGids.Add(g);
                        if (mappings.ContainsKey(g))
                        {
                            long ngid = g;
                            do
                            {
                                ngid = mappings[ngid];
                                mappedGids.Add(ngid);
                            } while (mappings.ContainsKey(ngid));
                        }
                    }
                    long newGid = mappedGids.Min();
                    foreach (long g in mappedGids)
                    {
                        if (g > newGid)
                            mappings[g] = newGid;
                    }
                } // if all of them are in a single groups, no need to store mapping
            }

            Console.WriteLine("{0}: first phase done", DateTime.Now);

            long[] allGids = mappings.Keys.ToArray();
            Stack<long> path = new Stack<long>();
            foreach (long g in allGids)
            {
                long ngid = mappings[g];
                while ((mappings.ContainsKey(ngid)))
                {
                    path.Push(ngid);
                    ngid = mappings[ngid];
                }
                while (path.Count > 0)
                { // remap for future
                    mappings[path.Pop()] = ngid;
                }
                ++k;
                yield return new Pair<long, long>(g, ngid);
            }

            Console.WriteLine("{0}: Total {1} new mappings were identified", DateTime.Now, k);
        }

        [Resource(IsStateful = true)]
        public static IEnumerable<Pair<long, long>> ResolveFoF_Final_opt(IEnumerable<CParticle> particles)
        {
            Console.WriteLine("{0}: load begin...", DateTime.Now);
            CParticle[] px = particles.ToArray(); // loaded in memory

            CParticle[][] groups = px.GroupBy((p) => p.Group).Select((g) => g.ToArray()).OrderBy((g) => -1 * g.Length).ToArray();

            Console.WriteLine("{0}: Loading done. {1} particles", DateTime.Now, px.Length);
            KDTreeOPT<CParticle> spIndex = new KDTreeOPT<CParticle>(ref px, 3, (p) => p.Coordinate);

            spIndex.IsRevisit = (p) => p.GridID != 0;
            spIndex.DoVisit = (p) => { p.GridID = -1; };

            int k = 0;

            for (int i = 0; i < px.Length; ++i)
            {
                px[i].GridID = 0;
            }

            Dictionary<long, long> mappings = new Dictionary<long, long>();

            Stack<CParticle> stack = new Stack<CParticle>();
            List<CParticle> buf = new List<CParticle>();
            float[] lowk = new float[3];
            float[] highk = new float[3];

            HashSet<long> gids = new HashSet<long>();
            HashSet<long> mappedGids = new HashSet<long>();

            for (int i = 0; i < groups.Length; ++i)
            {
                CParticle[] group = groups[i];
                gids.Clear();
                gids.Add(group[0].Group);
                for (int j = 0; j < group.Length; ++j)
                {
                    CParticle seed = group[j];
                    if (seed.GridID > 0) continue;

                    // recursively search
                    stack.Push(seed);

                    int groupSize = 1; // counting seed

                    while (stack.Count > 0)
                    {
                        CParticle p = stack.Pop();
                        if (p.GridID > 0) continue;
                        p.GridID = 1;// mark visited

                        buf.Clear();
                        spIndex.range(p, Radius, Period, (x) => p.distance(x, Period) <= Radius, buf);

                        foreach (CParticle x in buf)
                        {
                            stack.Push(x);
                            gids.Add(x.Group);
                        }

                        groupSize += buf.Count;
                    }
                }

                // now we got all neighboring groups
                // find all mapped groups

                if (gids.Count > 1)
                {
                    mappedGids.Clear();
                    foreach (long g in gids)
                    {
                        mappedGids.Add(g);
                        if (mappings.ContainsKey(g))
                        {
                            long ngid = g;
                            do
                            {
                                ngid = mappings[ngid];
                                mappedGids.Add(ngid);
                            } while (mappings.ContainsKey(ngid));
                        }
                    }
                    long newGid = mappedGids.Min();
                    foreach (long g in mappedGids)
                    {
                        if (g > newGid)
                            mappings[g] = newGid;
                    }
                } // if all of them are in a single groups, no need to store mapping
            }

            Console.WriteLine("{0}: first phase done", DateTime.Now);

            long[] allGids = mappings.Keys.ToArray();
            Stack<long> path = new Stack<long>();
            foreach (long g in allGids)
            {
                long ngid = mappings[g];
                while ((mappings.ContainsKey(ngid)))
                {
                    path.Push(ngid);
                    ngid = mappings[ngid];
                }
                while (path.Count > 0)
                { // remap for future
                    mappings[path.Pop()] = ngid;
                }
                ++k;
                yield return new Pair<long, long>(g, ngid);
            }

            Console.WriteLine("{0}: Total {1} new mappings were identified", DateTime.Now, k);
        }
        #endregion

        [Resource(IsStateful = true)]
        public static IEnumerable<Pair<long, long>> mergeGroup(IEnumerable<Pair<long,long>> mappings)
        {
            IOrderedEnumerable<Pair<long, long>> i = new List<Pair<long, long>>(mappings).OrderBy((x)=>x.Key);

            //Console.WriteLine("----Merging group----");
            Dictionary<long, long> mapping = new Dictionary<long, long>();
            long g;

            foreach ( Pair<long,long> p in i ) {
                if (mapping.ContainsKey(p.Value))
                {
                    g = mapping[p.Value];
                }
                else
                {
                    g = p.Value;
                }
                mapping[p.Key] = g;
                //Console.WriteLine("{0} > {1}", p.Key, g);
                yield return new Pair<long, long>(p.Key, g);
            }
        }

        #region non-uniform gridding
        static float PartitionAxis(ref CParticle[] src, Random rand, PartitionSpec spec) {
            float[] med = Util.FindMedian(ref src, spec.Start, spec.End, rand, spec.ValueSelector, spec.LessThanEqualTo);
            return med.Average();
        }

        class AxisComparer : IComparer<CParticle>
        {
            Util.ValueSelector<CParticle, float> m_vs;

            public AxisComparer(Util.ValueSelector<CParticle,float> vs)
            {
                m_vs = vs;
            }

            #region IComparer<CParticle> Members

            public int Compare(CParticle x, CParticle y)
            {
                float cx = m_vs(x);
                float cy = m_vs(y);
                return (cx == cy) ? 0 : ((cx < cy) ? -1 : 1);
            }

            #endregion
        }
        
        static float PartitionAxis(ref CParticle[] src, PartitionSpec spec)
        {
            Array.Sort(src, spec.Start, spec.Length, new AxisComparer(spec.ValueSelector));
            float avg = spec.ValueSelector(src[ spec.Middle1] );
            if ( (spec.Length & 0x01) == 0 ) {
                avg = ( avg + spec.ValueSelector(src[ spec.Middle2] ) ) * 0.5f;   
            }
            return avg;
        }

        class PartitionSpec {
            Util.ValueSelector<CParticle,float> valueSel;
            Util.LessThanEqualTo<float> le;
            int start;
            int end;

            public PartitionSpec(Util.ValueSelector<CParticle,float> vs,int s,int e) {
                valueSel = vs;
                le = ( x,y ) => x <= y;
                start = s;
                end = e;
            }

            public PartitionSpec(
                Util.ValueSelector<CParticle,float> vs,
                Util.LessThanEqualTo<float> op,int s, int e)
            {
                valueSel = vs; le = op; start = s; end = e;
            }

            public PartitionSpec(PartitionSpec p,int s,int e) {
                valueSel = p.valueSel;
                le = p.le;
                start = s;
                end = e;
            }

            public PartitionSpec[] split() {
                return split(valueSel);
            }

            public PartitionSpec[] split(Util.ValueSelector<CParticle,float> op) {
                return new PartitionSpec[] {
                    new PartitionSpec(op,start,Middle1), new PartitionSpec(op,Middle1+1,end)
                };
            }

            public int Start { get { return start; } }
            public int End { get { return end; } }
            public int Length { get { return end - start + 1; } }
            public int Middle1 { get { return (end + start) >> 1; } }
            public int Middle2 { get { return (end + start + 1) >> 1; } }
            public Util.ValueSelector<CParticle,float> ValueSelector { get { return valueSel; } }
            public Util.LessThanEqualTo<float> LessThanEqualTo { get { return le; } }
        };

        static GridTree Partition(int depth, float R, ref CParticle[] src)
        {
            List<Dictionary<long, Grid>> grids = new List<Dictionary<long, Grid>>();
            for (int i = 0; i <= depth; ++i)
                grids.Add(new Dictionary<long, Grid>());

            //GridTree.GridTreeNode root = new GridTree.GridTreeNode(new Grid(0, 0.0f, 0.0f, 0.0f, 1.0f, 1.0f, 1.0f));
            GridTree.GridTreeNode root = new GridTree.GridTreeNode(new Grid(0, -0.5f, -0.5f, -0.5f, 0.5f, 0.5f, 0.5f));
            PartitionSpec spec = new PartitionSpec((x) => x.X, 0, src.Length - 1);
            Partition(depth, root, ref src, spec, new Random(), grids);

            return new GridTree(root, grids, depth, R); 
        }

        private static GridTree.GridTreeNode Partition(int depth, GridTree.GridTreeNode g, ref CParticle[] src, PartitionSpec spec, Random rand,List<Dictionary<long,Grid>> grids)
        {
            grids[depth].Add(g.GridSpec.ID, g.GridSpec);

            if (depth > 0)
            {
                //float medX = PartitionAxis(ref src, rand, spec);
                float medX = PartitionAxis(ref src, spec);

                PartitionSpec[] sy = spec.split((x) => x.Y);
                float[] medY = { 
                        //PartitionAxis(ref src,rand,sy[0]),
                        //PartitionAxis(ref src,rand,sy[1]) 
                        PartitionAxis(ref src,sy[0]),
                        PartitionAxis(ref src,sy[1]) 
                };

                PartitionSpec[] sz1 = sy[0].split((x) => x.Z);
                PartitionSpec[] sz2 = sy[1].split((x) => x.Z);
                float[] medZ = {
                        //PartitionAxis(ref src, rand, sz1[0]),
                        //PartitionAxis(ref src, rand, sz1[1]),
                        //PartitionAxis(ref src, rand, sz2[0]),
                        //PartitionAxis(ref src, rand, sz2[1])
                        PartitionAxis(ref src, sz1[0]),
                        PartitionAxis(ref src, sz1[1]),
                        PartitionAxis(ref src, sz2[0]),
                        PartitionAxis(ref src, sz2[1])
                };

                GridTree.GridTreeNode[] subG = g.Split(medX, medY, medZ);

                List<PartitionSpec> specs = new List<PartitionSpec>();
                specs.AddRange(sz1[0].split((x) => x.X));
                specs.AddRange(sz1[1].split((x) => x.X));
                specs.AddRange(sz2[0].split((x) => x.X));
                specs.AddRange(sz2[1].split((x) => x.X));

                for (int i = 0; i < 8; ++i)
                {
                    Partition(depth - 1, subG[i], ref src, specs[i], rand, grids);
                }
            }

            return g;
        }
        #endregion

        #region Uniform partitioning
        static GridTree UniformPartition(int depth, float R)
        {
            List<Dictionary<long, Grid>> grids = new List<Dictionary<long, Grid>>();
            for (int i = 0; i <= depth; ++i)
                grids.Add(new Dictionary<long, Grid>());

            //GridTree.GridTreeNode root = new GridTree.GridTreeNode(new Grid(0, 0.0f, 0.0f, 0.0f, 1.0f, 1.0f, 1.0f));
            GridTree.GridTreeNode root = new GridTree.GridTreeNode(new Grid(0, -0.5f, -0.5f, -0.5f, 0.5f, 0.5f, 0.5f));
            root.Split(depth,grids);
            return new GridTree(root, grids, depth, R);
        }
        #endregion

        #region sampling operation
        [Homomorphic]
        [Associative]
        [Resource(IsStateful = false)]
        public static IEnumerable<T> randomSample<T>(IEnumerable<T> source, double rate)
        {
            Random rand = new Random();
            IEnumerator<T> i = source.GetEnumerator();
            while (i.MoveNext())
            {
                double r = rand.NextDouble();
                if (r < rate)
                    yield return i.Current;
            }
        }
        #endregion

        static GridTree computeGrid(IQueryable<CParticle> src,int depth,float R)
        {
            CParticle[] samples = src.ToArray();
            Console.WriteLine("{0} samples returned...", samples.Length);
            return Partition(depth,R,ref samples);
        }

        public static Expression<Func<CParticle,bool>> buildExprIsBoundary(GridTree t,int lv) {
            ParameterExpression p0 = Expression.Parameter(typeof(CParticle), "x");
            ConstantExpression obj = Expression.Constant(t);
            ConstantExpression level = Expression.Constant(lv);
            
            Type type = typeof(GridTree);
            MethodInfo mi = type.GetMethod("isBoundary", new Type[] { typeof(CParticle), typeof(int) });

            MethodCallExpression method = Expression.Call(obj,mi,new Expression[] { p0, level });
            
            return Expression.Lambda<Func<CParticle,bool>>(method, new ParameterExpression[] { p0 });
        }

        static void loadBinary(string file,List<CParticle> particles)
        {
            using (BinaryReader rd = new BinaryReader(new GZipStream(new FileStream(file, FileMode.Open, FileAccess.Read),CompressionMode.Decompress)))
            {
                try
                {
                    while (true)
                    {
                        particles.Add(CParticle.Read(rd));
                    }
                }
                catch (EndOfStreamException eof)
                {}

            }
        }

        static IQueryable<CParticle> loadParticles(string dataHost,int snapshot)
        {
            DryadDataContext ddc = new DryadDataContext(@"file://\\"+dataHost+@"\dryaddata\cosmo25cmb");

            IQueryable<BGasParticle> gas = ddc.GetPartitionedTable<BGasParticle>(String.Format("cosmo25cmb.{0:00000}_gas.bin.gz.pf",snapshot),CompressionScheme.GZipFast);
            IQueryable<BDarkParticle> dark = ddc.GetPartitionedTable<BDarkParticle>(String.Format("cosmo25cmb.{0:00000}_dark.bin.gz.pf", snapshot), CompressionScheme.GZipFast);
            IQueryable<BStarParticle> star = ddc.GetTable<BStarParticle>(String.Format("cosmo25cmb.768g.{0:00000}_star.bin",snapshot));

            IQueryable<CParticle> cgas = gas.Select((x) => x.toCParticle());
            IQueryable<CParticle> cdark = dark.Select((x) => x.toCParticle());
            IQueryable<CParticle> cstar = star.Select((x) => x.toCParticle());

            return cgas.Concat(cdark).Concat(cstar);
        }

        static IQueryable<CParticle> loadSubsampledParticles(string dataHost,int snapshot, int numNode,string subSample)
        {
            IQueryable<BGasParticle> gas = null;
            IQueryable<BDarkParticle> dark = null;
            IQueryable<BStarParticle> star = null;

            DryadDataContext ddc = new DryadDataContext(@"file://\\" + clusterName + @"\dryaddata");
            if (numNode > 1)
            {
                DryadDataContext ddc2 = new DryadDataContext(@"file://\\"+dataHost+@"\dryaddata\cosmo25cmb");
                gas = ddc2.GetPartitionedTable<BGasParticle>(String.Format("cosmo25cmb.{1}.{0:00000}_gas.bin.gz.pf", snapshot, subSample));
                dark = ddc2.GetPartitionedTable<BDarkParticle>(String.Format("cosmo25cmb.{1}.{0:00000}_dark.bin.gz.pf", snapshot, subSample));
            }
            else
            {
                gas = ddc.GetTable<BGasParticle>(String.Format("cosmo25cmb.768g.{1}.{0:00000}_gas.bin.gz", snapshot, subSample), CompressionScheme.GZipFast);
                dark = ddc.GetTable<BDarkParticle>(String.Format("cosmo25cmb.768g.{1}.{0:00000}_dark.bin.gz", snapshot, subSample), CompressionScheme.GZipFast);
            }

            star = ddc.GetTable<BStarParticle>(String.Format("cosmo25cmb.768g.{1}.{0:00000}_star.bin", snapshot, subSample));

            IQueryable<CParticle> cgas = gas.Select((x) => x.toCParticle());
            IQueryable<CParticle> cdark = dark.Select((x) => x.toCParticle());
            IQueryable<CParticle> cstar = star.Select((x) => x.toCParticle());

            return cgas.Concat(cdark).Concat(cstar);
        }

        static IQueryable<CParticle> loadParticles(string dataHost,int snapshot,int numNode)
        {
            IQueryable<BGasParticle> gas = null;
            IQueryable<BDarkParticle> dark = null;
            IQueryable<BStarParticle> star = null;

            DryadDataContext ddc = new DryadDataContext(@"file://\\" + clusterName + @"\dryaddata");
            if (numNode > 1)
            {
                DryadDataContext ddc2 = new DryadDataContext(@"file://\\"+dataHost+@"\dryaddata\cosmo25cmb");
                gas = ddc2.GetPartitionedTable<BGasParticle>(String.Format("cosmo25cmb.{0:00000}_gas.{1}.bin.gz.pf", snapshot, numNode), CompressionScheme.GZipFast);
                dark = ddc2.GetPartitionedTable<BDarkParticle>(String.Format("cosmo25cmb.{0:00000}_dark.{1}.bin.gz.pf", snapshot, numNode), CompressionScheme.GZipFast);
            }
            else
            {
                gas = ddc.GetTable<BGasParticle>(String.Format("cosmo25cmb.768g.{0:00000}_gas.bin.gz", snapshot), CompressionScheme.GZipFast);
                dark = ddc.GetTable<BDarkParticle>(String.Format("cosmo25cmb.768g.{0:00000}_dark.bin.gz", snapshot), CompressionScheme.GZipFast);
            }

            star = ddc.GetTable<BStarParticle>(String.Format("cosmo25cmb.768g.{0:00000}_star.bin", snapshot));

            IQueryable<CParticle> cgas = gas.Select((x) => x.toCParticle());
            IQueryable<CParticle> cdark = dark.Select((x) => x.toCParticle());
            IQueryable<CParticle> cstar = star.Select((x) => x.toCParticle());

            return cgas.Concat(cdark).Concat(cstar);
        }

        static IQueryable<CParticle> loadParticles(string dataHost,int snapshot,GridTree gridTree)
        {
            DryadDataContext ddc = new DryadDataContext(@"file://\\"+dataHost+@"\dryaddata\cosmo25cmb");

            IQueryable<BGasParticle> gas = ddc.GetPartitionedTable<BGasParticle>(String.Format("cosmo25cmb.{0:00000}_gas.bin.gz.pf", snapshot), CompressionScheme.GZipFast);
            IQueryable<BDarkParticle> dark = ddc.GetPartitionedTable<BDarkParticle>(String.Format("cosmo25cmb.{0:00000}_dark.bin.gz.pf", snapshot), CompressionScheme.GZipFast);
            IQueryable<BStarParticle> star = ddc.GetTable<BStarParticle>(String.Format("cosmo25cmb.768g.{0:00000}_star.bin", snapshot));

            // FIXME: two consecutive select across concat causes error
            IQueryable<CParticle> cgas = gas.Select((x) => x.toCParticle()).Select((x) => new CParticle(gridTree.GetGridID(x), 0, x));
            IQueryable<CParticle> cdark = dark.Select((x) => x.toCParticle()).Select((x) => new CParticle(gridTree.GetGridID(x), 0, x));
            IQueryable<CParticle> cstar = star.Select((x) => x.toCParticle()).Select((x) => new CParticle(gridTree.GetGridID(x), 0, x));

            return cgas.Concat(cdark).Concat(cstar);
        }

        [Serializable]
        struct LongComparer : IComparer<long>, IEqualityComparer<Pair<long,long>>
        {

            #region IComparer<long> Members

            public int Compare(long x, long y)
            {
                return (int)(x - y);
            }

            #endregion


            #region IEqualityComparer<Pair<long,long>> Members

            public bool Equals(Pair<long, long> x, Pair<long, long> y)
            {
                return x.Key == y.Key && x.Value == y.Value;
            }

            public int GetHashCode(Pair<long, long> obj)
            {
                return (int)obj.Key;
            }

            #endregion
        }

        #region Global Parameters
        public static float Period = 1.0f;

        public static float Radius = 0.00026042f;
        public static int   MinParticles = 16;

        public static bool UniformGrid = false;
        public static bool KDTreeNaive = false;
        #endregion

        static string clusterName = null;
        static string subSample = null;
        static string dataHost = null;

        static void Main(string[] args)
        {
            int snapshot = 43;
            int numNode = 0;
            int maxLevel = 2;
            int i = 0;

            while (i < args.Length)
            {
                if ("-s".Equals(args[i]))
                {
                    snapshot = Int32.Parse(args[++i]);
                }
                else if ("-host".Equals(args[i]))
                {
                    dataHost = args[++i];
                }
                else if ("-u".Equals(args[i]))
                {
                    UniformGrid = true;
                }
                else if ("-n".Equals(args[i]))
                {
                    KDTreeNaive = true;
                }
                else if ("-minpts".Equals(args[i]))
                {
                    MinParticles = Int32.Parse(args[++i]);
                }
                else if ("-level".Equals(args[i]))
                {
                    maxLevel = Int32.Parse(args[++i]);
                }
                else if ("-linklen".Equals(args[i]))
                {
                    Radius = float.Parse(args[++i]);
                }
                else if ("-node".Equals(args[i]))
                {
                    numNode = Int32.Parse(args[++i]);
                }
                else if ("-quiet".Equals(args[i]))
                {
                    DryadLinq.Verbose = false;
                }
                else if ("-visual".Equals(args[i]))
                {
                    DryadLinq.Visualization = args[++i];
                }
                else if ("-cluster".Equals(args[i]))
                {
                    clusterName = args[++i];
                }
                else if ("-subSample".Equals(args[i]))
                {
                    subSample = args[++i];
                }
                else
                {
                    Console.Error.WriteLine("Invalid option: {0}", args[i]);
                }
                ++i;
            }

            if ( clusterName != null ) {
                if (numNode > 1)
                {
                    DryadLinq.ClusterName = String.Format("{0}{1}", clusterName, numNode);
                }
                else if (numNode == 1)
                {
                    DryadLinq.ClusterName = clusterName;
                }
                else if (numNode < 0)
                {
                    DryadLinq.ClusterName = String.Format("{0}{1}", clusterName, numNode * -1);
                    numNode = 1;
                }
            }

            Console.WriteLine("Running processid = {0}", Process.GetCurrentProcess().Id);

            Console.WriteLine("Using cluster {0}", DryadLinq.ClusterName);
            
            Console.WriteLine("Snapshot = {0}", snapshot);
            Console.WriteLine("Number of Nodes = {0}", numNode);
            Console.WriteLine("Uniform Partition = {0}", UniformGrid);
            Console.WriteLine("Naive KDTree = {0}", KDTreeNaive);
            Console.WriteLine("Minimum Points = {0}", MinParticles);
            Console.WriteLine("Linking Length = {0}", Radius);

            Run(dataHost,snapshot,numNode,maxLevel);

            Console.ReadKey();
        }

        static void Run(string dataHost,int snapshot,int numNode,int maxLevel)
        {
            int level = 0;

            using (StreamWriter log = new StreamWriter(String.Format("progress_{0}.txt",DryadLinq.ClusterName), true))
            {
                log.AutoFlush = true;
                log.WriteLine("{0}: ----- BEGIN -----", DateTime.Now);
                
                GridTree gridTree = null;
                IQueryable<CParticle> gridAndParticles = null;
                long[] buckets = null;

                if (UniformGrid)
                {
                    gridTree = UniformPartition(maxLevel, Radius);
                    buckets = gridTree.toBuckets(level);
                    gridAndParticles = loadParticles(dataHost,snapshot, gridTree);
                }
                else
                {
                    // for real
                    IQueryable<CParticle> cparticles = null;
                    if ( numNode == 0 ) {
                        cparticles = loadParticles(dataHost,snapshot);
                    } else if ( subSample == null ) {
                        cparticles = loadParticles(dataHost, snapshot, numNode);
                    } else {
                        cparticles = loadSubsampledParticles(dataHost, snapshot, numNode, subSample);
                    }
                    
                    IQueryable<CParticle> sparticles = cparticles.Apply((x) => randomSample(x, 0.001)).ToDryadTableLazy();
                    
                    IQueryable<CParticle> particles = cparticles.ToDryadPartitionedTableLazy();

                    // compute grids from sample
                    log.WriteLine("{0}: sampling start...", DateTime.Now);

                    IDryadTable[] mresult = DryadLinqQueryable.Materialize(sparticles, particles);
                    sparticles = mresult[0] as IQueryable<CParticle>;
                    particles = mresult[1] as IQueryable<CParticle>;

                    // run both initialization and sampling at the same time
                    gridTree = computeGrid(sparticles, maxLevel, Radius);

                    buckets = gridTree.toBuckets(level);

                    gridAndParticles = particles.Select((x) => new CParticle(gridTree.GetGridID(x), 0, x));
                }

                log.WriteLine("{0}: begin initial clustering!", DateTime.Now);

                IQueryable<Pair<long, long>> allMap = null;

                //IQueryable<CParticle> gridAndParticles = particles.Select((x) => new CParticle(gridTree.GetGridID(x), 0, x));
                IQueryable<CParticle> partitionedInGrids = gridAndParticles.RangePartition((p) => p.GridID, buckets, new LongComparer());
                IQueryable<CParticle> initClusters = ( KDTreeNaive ) ? partitionedInGrids.Apply((x) => fof_naive(x)) : partitionedInGrids.Apply((x) => fof_opt(x));
                IQueryable<CParticle> boundary = initClusters.Where( buildExprIsBoundary(gridTree,level) );

                IQueryable<CParticle> initialClusters = initClusters;
                IQueryable<CParticle> nextBoundary = boundary;
                IQueryable<CParticle> partitions = null;

                while ( level < maxLevel )
                {
                    ++level;

                    // now we'll do range partitioning on grids
                    buckets = gridTree.toBuckets(level);

                    if (buckets.Length > 1)
                    {
                        partitions = nextBoundary.RangePartition((x) => x.GridID, buckets, new LongComparer());
                    }
                    else
                    {
                        partitions = nextBoundary.Merge();
                    }
                    
                    IQueryable<Pair<long, long>> mappings = ( KDTreeNaive ) ? partitions.Apply((x) => ResolveFoF_naive(x)) : partitions.Apply((x) => ResolveFoF_opt(x));

                    if (allMap == null)
                    {
                        allMap = mappings;
                    }
                    else
                    {
                        allMap = allMap.Union(mappings, new LongComparer());
                    }

                    // should prepare next iteration
                    boundary = partitions.Where( buildExprIsBoundary(gridTree,level) );
                    nextBoundary = boundary.GroupJoin(mappings, (o) => o.Group, (i) => i.Key, (o, i) => ( i.Count() == 0 ) ? o : new CParticle(o.GridID, i.Single().Value, o));

                    log.WriteLine("{0}: iteration {1} completed!", DateTime.Now, level);
                }

                log.WriteLine("{0}: while loop completed!", DateTime.Now);

                if (allMap == null)
                {
                    //allMap = nextBoundary.Apply((p) => ResolveFoF_Final2(p));
                    allMap = ( KDTreeNaive ) ? nextBoundary.Apply((p) => ResolveFoF_Final_naive(p)) : nextBoundary.Apply((p) => ResolveFoF_Final_opt(p));
                }
                else
                {
                    IQueryable<Pair<long, long>> finalMap = ( KDTreeNaive ) ? nextBoundary.Apply((p) => ResolveFoF_Final_naive(p)) : nextBoundary.Apply((p) => ResolveFoF_Final_opt(p));
                    allMap = allMap.Union(finalMap,new LongComparer());
                }

                // let's do relabeling
                IQueryable<Pair<long, long>> allMapMerge = allMap.Merge();

                // order by returns zero length stream
                IQueryable<Pair<long, long>> finalMapping = allMapMerge.Apply((x) => mergeGroup(x));
                IQueryable<Pair<long, long>> finalResult = initialClusters.GroupJoin(finalMapping, (o) => o.Group, (i) => i.Key, (o, i) => new Pair<long, long>(o.ParticleID, (i.Count() == 0) ? o.Group : i.Single().Value));
                
                // prune
                IQueryable<Pair<long, long>> groupSize = finalResult.Select((p) => p.Value).GroupBy((g) => g).Select((g) => new Pair<long, long>(g.Key, g.Count()));
                IQueryable<Pair<long, long>> closeToFinal = finalResult.Join(groupSize, (o) => o.Value, (i) => i.Key, (o, i) => (i.Value < MinParticles) ? new Pair<long, long>(o.Key, 0) : o);
                IQueryable<Pair<long, long>> realClusters = closeToFinal.Where((x) => x.Value > 0);

                realClusters.Join(initialClusters, (o) => o.Key, (i) => i.ParticleID, (o, i) => new CParticle(0, o.Value, i)).ToDryadPartitionedTable(String.Format(@"file://\\"+dataHost+@"\dryaddata\cluster-group-{0}-{1}.pf",DryadLinq.ClusterName,snapshot));

                log.WriteLine("{0}: done merging!", DateTime.Now);
            }
        }
    }
}
